import sys
import os.path
import pickle
import random
import pygame
import pygame.mixer
import threading
import time
import string
import cgi
from os import curdir, sep

#For Windows Compatability
#from http.server import BaseHTTPRequestHandler, HTTPServer
from BaseHTTPServer import BaseHTTPRequestHandler, HTTPServer

# Global variables
#For Windows Compatability
#PATH_SEPERATOR = "\\"
#MUSIC_PATH = ".\\music\\"
PATH_SEPERATOR = "/"
#MUSIC_PATH = "./music/"
#MUSIC_PATH = "x:\\"
MUSIC_PATH = "/mnt/mediaserver/FamilyMusic/"
catelogue = None
playList = []
isAlbumMode = False
currentlyPlaying = None
keepAlive = True
childFriendlyOnly = False
masterVolume = 10

def showMenu():
    global playList
    
    choice = None
    while choice != "9":
        print("CTRL-C For Menu")
        try:
            while 1:
                time.sleep(1)
        except KeyboardInterrupt:
            print("Menu")
        print(
            """
            Blueberry Music
    
            0 - Rebuild database
            1 - Build playlist
            2 - Show status
            3 - Toggle Album Mode
            5 - Play Next Track
            6 - Pause playback
            7 - Resume playback
            8 - Pickle data
            9 - Quit
            """
            )
        print("Quick Quit Active")
        #Cast to Sring not necessary in Windows
        choice = str(input("Select action: "))
        print()
        if choice == "0":
            buildDatabase()
        elif choice == "1":
            playList = []
            resetPlayList()
        elif choice == "2":
            showStatus()
        elif choice == "3":
            toggleAlbumMode()
        elif choice == "5":
            playNext()
        elif choice == "6":
            pausePlayback()
        elif choice == "7":
            resumePlayback()
        elif choice == "8":
            pickleCatelogue()
        elif choice == "9":
            print("Menu Quitting")
        else:
            print("Invalid selection")

def toggleAlbumMode():
    global isAlbumMode
    global playList
    
    if isAlbumMode:
        print("Switching to Single Mode")
        isAlbumMode = False
    else:
        print("Switching to Album Mode")
        playList = []
        isAlbumMode = True
    resetPlayList()


def setAlbumMode():
    global isAlbumMode
    global playList
    print("Switching to Album Mode")
    playList = []
    isAlbumMode = True
    resetPlayList()


def setSingleMode():
    global isAlbumMode
    global playList
    print("Switching to Single Mode")
    isAlbumMode = False
    resetPlayList()

def buildDatabase():
    print("Building database from: " + MUSIC_PATH)
    global catelogue
    try:
        myFile = open(MUSIC_PATH + "prefdata.dat", "rb")
        catelogue = pickle.load(myFile)
        myFile.close()
    except IOError as e:
        print("New pickle jar created")
        catelogue = {}

    # Used to store the file names to check for disappearences
    fileList = []
    trackList = ()
    for files in os.walk(MUSIC_PATH) :
        if (not files[1]) : #Files[1] is the directory
            lastTrack = None #Each track is added on the sucessive loop, so it knows the name of the next track to play
            trackList = sorted(files[2])
            for tracks in trackList : #Loops for all the tracks within a directory
                if (tracks.find(".ogg") > 2) : # Edits out all the album art and other rubbish
                    fileList.append(files[0] + PATH_SEPERATOR + tracks)
                    if (lastTrack) :
                        if lastTrack not in catelogue:
                            print("Track added:", lastTrack[lastTrack.rfind(PATH_SEPERATOR)+1:len(lastTrack)-4])
                            catelogue[lastTrack] = [60, 1, files[0] + PATH_SEPERATOR + tracks]
                    lastTrack = (files[0] + PATH_SEPERATOR + tracks)
            if lastTrack not in catelogue: #Last track of each album is added here
                print("Track added:", lastTrack[lastTrack.rfind(PATH_SEPERATOR)+1:len(lastTrack)-4])
                catelogue[lastTrack] = [60, 1, "None"];

    # This loop makes a list of songs in the catelogue which don't have corresponding files
    toBeCleared = []
    for song in catelogue :
        if (song not in fileList) :
            toBeCleared.append(song)
    # This loop deletes the cateloge entires identified above
    for song in toBeCleared :
        print("Marked for removal:", song)
        del catelogue[song]

def pickleCatelogue():
    print("Catelogue being pickled")
    myFile = open(MUSIC_PATH + "prefdata.dat", "wb")
    pickle.dump(catelogue, myFile)
    myFile.close();

def resetPlayList():
    global playList
    global currentlyPlaying
    global isAlbumMode

    if isAlbumMode:
        if len(playList) < 1:
            if catelogue[currentlyPlaying][2] != "None":
                playList.append(catelogue[currentlyPlaying][2])
            else:
                playList.append(weightedSelection())
            
        while len(playList) < 5:
            if catelogue[playList[len(playList)-1]][2] != "None":
                playList.append(catelogue[playList[len(playList)-1]][2])
            else:
                playList.append(weightedSelection())
                isAlbumMode = False
                break
    
    while len(playList) < 5:
        playList.append(weightedSelection())
        
def weightedSelection():
    global childFriendlyOnly
    global playList
    global currentlyPlaying
    
    selectedTrack = None
    if len(playList) < 1:
        selectedTrack = random.choice(list(catelogue.keys()))
    else:
        if catelogue[playList[len(playList)-1]][2] != "None":
            selectedTrack = catelogue[playList[len(playList)-1]][2]
        else:
            selectedTrack = random.choice(list(catelogue.keys()))

    while True:
        if int(catelogue[selectedTrack][0]) > random.randint(0,100):
            if childFriendlyOnly:
                if catelogue[selectedTrack][1] != 1:
                    print("Track rejected as not child friendly")
                else:
                    break
            elif "NoRandomPlay" in selectedTrack:
                print("Track rejected as spoken word")
            else:
                break
        else: #If the track isn't randomly selected we nudge up the Liked score to make it more likely to come up next time
            catelogue[selectedTrack][0] += 1
        selectedTrack = random.choice(list(catelogue.keys()))

    catelogue[selectedTrack][0] -= 1 #Before a selected track is returned we nudge down the score to make it less likely to come up again
    return selectedTrack

def showStatus():
    global currentlyPlaying
    print("Currently Playing: " + currentlyPlaying)
    print()
    showPlayList()

def showPlayList():
    global playList
    print("Queued Tracks")
    print()
    for eachTrack in playList:
        print(eachTrack[eachTrack.rfind(PATH_SEPERATOR)+1:len(eachTrack)-4])

def playTrack():
    global currentlyPlaying
    print("Lanching track: ", currentlyPlaying)
    setVolume()
    pygame.mixer.music.load(currentlyPlaying)
    pygame.mixer.music.play()

def playNext():
    global playList
    global currentlyPlaying
    currentlyPlaying = playList.pop(0)
    resetPlayList()
    playTrack()

def pausePlayback():
    pygame.mixer.music.pause()

def resumePlayback():
    pygame.mixer.music.unpause()

def restartPlayback():
    pygame.mixer.music.rewind()

def setVolume():
    global masterVolume
    pygame.mixer.music.set_volume(masterVolume/10)

def volumeUp():
    global masterVolume
    if masterVolume < 10:
        masterVolume = masterVolume + 1
    setVolume()

def volumeDown():
    global masterVolume
    if masterVolume > 0:
        masterVolume = masterVolume - 1
    setVolume()
    
def trackEnded():
    print("That's the end of a track")

def ears ():
    global keepAlive
    while keepAlive:
        if not pygame.mixer.music.get_busy():
            playNext()
        time.sleep(.1)

def startTheMusic():
    #pygame.mixer.pre_init(44100,-16,2,4096)
    #pygame.mixer.init(44100,-16,2,4096)
    pygame.mixer.pre_init(44100,-16,2,2048)
    pygame.mixer.init(44100,-16,2,2048)
    setVolume()
    buildDatabase()
    resetPlayList()
    playNext()

class MyHandler(BaseHTTPRequestHandler):
    global currentlyPlaying
    global playList
    global isAlbumMode
    global childFriendlyOnly
    
    def do_GET(self):
        try:
            if self.path.endswith(".gif"):
                f = open(self.path[1:], "rb")
                self.send_response(200)
                self.send_header('Content-type', 'image')
                self.end_headers()
                self.wfile.write(f.read())
                f.close()
                return
            if self.path.endswith(".html"):
                f = open(curdir + sep + self.path)
                self.send_response(200)
                self.send_header('Content-type',	'text/html')
                self.end_headers()
                self.wfile.write(f.read().encode("utf-8"))
                f.close()
                return
            if self.path.endswith("conductor.esp"):   #our dynamic content
                self.send_response(200)
                self.send_header('Content-type',	'text/html')
                self.end_headers()
                s = self.__getStatus()
                self.wfile.write(s.encode("utf-8"))
                return
            if self.path.endswith(".cmd"):
                s = "Command Completed"
                if self.path.endswith("next.cmd"):
                    playNext()
                    s = "#Track Skipped"
                elif self.path.endswith("pause.cmd"):
                    pausePlayback()
                    s = "-Track Paused"
                elif self.path.endswith("play.cmd"):
                    resumePlayback()
                    s = "-Track Resumed"
                elif self.path.endswith("restart.cmd"):
                    restartPlayback()
                    s = "-Track Restarted"
                elif self.path.endswith("albumMode.cmd"):
                    setAlbumMode()
                    s = "#Resetting to album mode"
                elif self.path.endswith("singleMode.cmd"):
                    setSingleMode()
                    s = "#Resetting to single mode"
                elif self.path.endswith("prudeMode.cmd"):
                    print("Setting to prudish mode")
                    childFriendlyOnly = True
                    s = "#Resetting to child-friendly only"
                elif self.path.endswith("adultMode.cmd"):
                    print("Setting to aldult mode")
                    childFriendlyOnly = False
                    s = "#Resetting to anything-goes mode"
                elif self.path.endswith("volumeUp.cmd"):
                    volumeUp()
                    s = "-Volume Increased"
                elif self.path.endswith("volumeDown.cmd"):
                    volumeDown()
                    s = "-Volume Decreased"
                elif self.path.endswith("getArtistList.cmd"):
                    s = self.__artistList()
                elif self.path.endswith("getAlbumList.cmd"):
                    s = self.__substitute(self.path[1:len(self.path)-17])
                    s = self.__albumList(s)
                elif self.path.endswith("getTrackList.cmd"):
                    s = self.__substitute(self.path[1:len(self.path)-17])
                    s = self.__trackList(s)
                elif self.path.endswith("getTrackListWithDetails.cmd"):
                    s = self.__trackListWithDetails()
                    print(s)
                elif self.path.endswith("setTrack.cmd"):
                    s = self.__substitute(self.path[1:len(self.path)-13])
                    s = self.__setTrack(s)
                elif self.path.endswith("sendUpdatePlus.cmd"):
                    s = "Status Update Plus" + "\n" + currentlyPlaying[currentlyPlaying.rfind(PATH_SEPERATOR)+1:len(currentlyPlaying)-4] + "\n"
                    for song in playList:
                        s = s + song[song.rfind(PATH_SEPERATOR)+1:len(song)-4] + "\n"
                    #This sends back a status, but triggers a second requst from the client for the album listing
                elif self.path.endswith("sendUpdate.cmd"):
                    s = "Status Update" + "\n" + currentlyPlaying[currentlyPlaying.rfind(PATH_SEPERATOR)+1:len(currentlyPlaying)-4] + "\n"
                    for song in playList:
                        s = s + song[song.rfind(PATH_SEPERATOR)+1:len(song)-4] + "\n"
                elif self.path.endswith("processChange.cmd"):
                    s = self.__substitute(self.path[1:len(self.path)-17])
                    s = self.__processChange(s)
                    s = "Status Update" + "\n" + currentlyPlaying[currentlyPlaying.rfind(PATH_SEPERATOR)+1:len(currentlyPlaying)-4] + "\n"
                    for song in playList:
                        s = s + song[song.rfind(PATH_SEPERATOR)+1:len(song)-4] + "\n"
                    
                self.send_response(200)
                self.send_header('Content-type',	'text/html')
                self.end_headers()
                self.wfile.write(s.encode("utf-8"))
                return
            return
                
        except IOError:
            self.send_error(404,'File Not Found: %s' % self.path)

    def __processChange(self, rawString):
        print("Change Received")
        processAlbumName = rawString[:rawString.find("-")-1] + PATH_SEPERATOR + rawString[rawString.find("-")+2:rawString.find(":")]
        processString = rawString[rawString.find(":")+1:]
        processTrack = ""
        while len(processString) > 3:
            processTrack = MUSIC_PATH + processAlbumName + PATH_SEPERATOR + processString[:processString.find(":")] + ".ogg"
            processString = processString[processString.find(":")+1:]
            catelogue[processTrack][0] = int(processString[:processString.find(":")])
            processString = processString[processString.find(":")+1:]
            catelogue[processTrack][1] = int(processString[:processString.find(":")])
            processString = processString[processString.find(":")+1:]
        pickleCatelogue()
            
    def __substitute(self, rawString):
        return rawString.replace("%20", " ")

    def __setTrack(self, rawTrack):
        print(rawTrack)
        position = int(rawTrack[len(rawTrack)-1])
        track = rawTrack[0:len(rawTrack)-2]
        album = track[track.find(":")+1:track.rfind(":")]
        track = track[track.rfind(":")+1:len(track)] + ".ogg"
        band = rawTrack[0:rawTrack.find(":")]
        searchString = MUSIC_PATH + band + PATH_SEPERATOR + album
        for song in list(catelogue.keys()):
            if searchString in song:
                if track in song:
                    playList[position-1] = song
        return "#Track set"

    def __trackList(self, artistAlbum):
        trackList = []
        searchString = MUSIC_PATH + artistAlbum.replace(":", PATH_SEPERATOR);
        returnValue = "Track List\n"
        for song in list(catelogue.keys()):
            if searchString in song:
                trackList.append(song)
        for song in sorted(trackList):
            returnValue = returnValue + song[song.find("-")+2:len(song)-4] + "\n"
        return returnValue
    
    def __trackListWithDetails(self):
        trackList = []
        searchString = currentlyPlaying[:currentlyPlaying.rfind(PATH_SEPERATOR)];
        lineOne = searchString[len(MUSIC_PATH):]
        returnValue = "Track Details\n"
        returnValue = returnValue + lineOne[:lineOne.rfind(PATH_SEPERATOR)] + " - " + lineOne[lineOne.rfind(PATH_SEPERATOR)+1:] + "\n"
        for song in list(catelogue.keys()):
            if searchString in song:
                trackList.append(song)
        for song in sorted(trackList):
            returnValue = returnValue + song[song.find("-")+2:len(song)-4] + ":" + str(catelogue[song][0]) + ":" + str(catelogue[song][1]) + "\n"
        return returnValue
    
    def __albumList(self, artist):
        print("Finding albums matching " + artist)
        rawsong = ""
        albumName = ""
        rawAlbums = []
        for song in list(catelogue.keys()):
            rawsong = (song[len(MUSIC_PATH):len(song)])
            if rawsong.startswith(artist):
                rawsong = rawsong[rawsong.find(PATH_SEPERATOR)+1:]
                rawsong = rawsong[0:rawsong.find(PATH_SEPERATOR)]
                if rawsong not in rawAlbums:
                    rawAlbums.append(rawsong)
        returnValue = "Album List\n"
        for song in rawAlbums:
            returnValue = returnValue + song + "\n"
        return returnValue
    
    def __artistList(self):
        rawlist = []
        rawsong = ""
        for song in list(catelogue.keys()):
            rawsong = (song[len(MUSIC_PATH):len(song)])
            rawsong = rawsong[0:rawsong.find(PATH_SEPERATOR)]
            if rawsong not in rawlist:
                rawlist.append(rawsong)
        rawlist.sort()
        returnValue = "Artist List\n"
        for song in rawlist:
            returnValue = returnValue + song + "\n"
        return returnValue

    def __getStatus(self):
        bandName = str(currentlyPlaying[currentlyPlaying.rfind(PATH_SEPERATOR)+4:currentlyPlaying.rfind("-")])
        trackName = str(currentlyPlaying[currentlyPlaying.rfind("-")+2:len(currentlyPlaying)-4])
        builtStatus = "<html><header><title>Python Conductor</title></header>"
        builtStatus = builtStatus + "<body><center><table border=5 cellpadding=4><tr><td>"
        builtStatus = builtStatus + bandName + "</td></tr><tr><td>" + trackName
        builtStatus = builtStatus + "</td></tr></table>"
        builtStatus = builtStatus + "</br>"
        builtStatus = builtStatus + "<table border=0><tr>"
        builtStatus = builtStatus + '<td><a href="restart.cmd"><img src="restart.gif"></a></td>'
        builtStatus = builtStatus + '<td><a href="pause.cmd"><img src="pause.gif"></a></td>'
        builtStatus = builtStatus + '<td><a href="play.cmd"><img src="play.gif"></a></td>'
        builtStatus = builtStatus + '<td><a href="next.cmd"><img src="next.gif"></a></td>'
        builtStatus = builtStatus + '</tr></table>'
        builtStatus = builtStatus + "</br>"
        builtStatus = builtStatus + "<table border=5 cellpadding=4>"
        for song in playList:
            builtStatus = builtStatus + "<tr><td>" + song[song.rfind(PATH_SEPERATOR)+1:len(song)-4] + "</td></tr>"
        builtStatus = builtStatus + "</table>"
        builtStatus = builtStatus + "</br>"
        builtStatus = builtStatus + "<table border=0><tr>"
        builtStatus = builtStatus + '<td><a href="albumMode.cmd"><img src="album.gif"></a></td>'
        builtStatus = builtStatus + '<td><a href="singleMode.cmd"><img src="single.gif"></a></td>'
        builtStatus = builtStatus + '<td><a href="volumeDown.cmd"><img src="vol_down.gif"></a></td>'
        builtStatus = builtStatus + '<td><a href="volumeUp.cmd"><img src="vol_up.gif"></a></td>'
        builtStatus = builtStatus + '</tr></table>'
        builtStatus = builtStatus + "</center></body>"
        return builtStatus

def webServer():       
    try:
        server = HTTPServer(('', 80), MyHandler)
        print ("started httpserver...")
        server.serve_forever()
    except:
        print ("Web server failed")


startTheMusic()

t = threading.Thread(target=ears)
t.start()

g = threading.Thread(target=webServer)
g.start()

showMenu()
pickleCatelogue()
keepAlive = False
pygame.mixer.quit()

